"""Energy Control Modules Add-ons
"""

from __future__ import print_function
from isotel.idm import gateway
import sys
import time
import argparse
import textwrap
from enum import IntFlag, IntEnum, unique

def _eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class ECL(gateway.Device):

    def __init__(self, group = None, name = None, exclude = None):
        super().__init__(group, name, exclude, unit=True, accuracy=False, advanced=True, development=True, hidden=True, brand_filter='ECL-')

    def set_time(self):
        return None

    def set_maxcurrents(self, I1, I2, I3):
        return None


class EC(gateway.Device):
    EC_LOG_BUF_LEN = 992
    HEADER_MASK = 0xF0
    HEADER_DATA_MASK = 15
    LOG_ID_ENERGY = 0x10
    LOG_ID_FAULT = 0x20
    LOG_ID_TIME_CHANGE = 0x30
    LOG_ID_SUBMETERING = 0x40
    LOG_ID_TRACKING = 0x50
    LOG_ID_ENERGY1 = 0x70
    LOG_REASON_SYSTEM_POWER_UP = 1
    LOG_REASON_WDT_RESET_OCCUR = 2
    LOG_REASON_TIME_CHANGE = 4

    @unique
    class ChannelStatus(IntEnum):
        OFF = 0
        ANTISMOG = 1
        ON = 2
        SHEDDED = 3
        LIMITED = 4
        CURRENT_FAULT = 5
        WAITING_PRIORITY = 6
        VOLTAGE_FAULT = 7
        UNDERVOLTAGE_FAULT = 8
        OVERVOLTAGE_FAULT = 9
        OVERUSE_FAULT   = 10

    @unique
    class CommonOpts(IntFlag):
        DETECT_CAPLOAD  = 0x02  # Enable capacitive load in load detection, as with anti-smog
        BOOST_25A       = 0x04  # Enable 25 A mode, only valid in single channel mode operation
        COMMON_OP       = 0x08  # Treat both A and B channels in a single channel mode
        AUTOLEARN       = 0x10  # Calculate average energy consumption
        CURRENT_FAULT   = 0x20  # Current Fault Enable, by default on
        FAST_OFF_EN	    = 0x40  # Enable short-circuit tripping, by default on
        SLOW_OFF        = 0x80  # Set slow reaction time on overloading, by default off

    @unique
    class ChannelOpts(IntFlag):
        OPTS_UNDER_V_FAULT	= 0x01  # Enable under voltage faults
        OPTS_ELIMIT_OVERRIDE= 0x02
        OPTS_ELIMIT_ENABLE	= 0x04
        OPTS_ASMOG_ENABLE	= 0x08  # Enable Anti-Smog
        OPTS_SINGLE_OP		= 0x10	
        OPTS_ELIMIT_CUT_OFF	= 0x20  # Enable Excessive Energy consumption Fault
        OPTS_GEN_ELIMIT_EN	= 0x40
        OPTS_EXT_INPUT		= 0x80  # Enable External Gate Input

    def __init__(self, group = None, name = None, exclude = None):
        super().__init__(group, name, exclude, unit=True, accuracy=False, advanced=True, development=True, hidden=True, brand_filter='EC-')

    def geta_status(self):
        return self.ChannelStatus(int(self.get_value('ec.ChA', after='now')))

    def getb_status(self):
        return self.ChannelStatus(int(self.get_value('ec.ChB', after='now')))

    def get_status(self):
        return [self.geta_status(), self.getb_status()]

    def seta(self, turn_on=True):
        return self.ChannelStatus(int(self.set_value('ec.ChA', int(self.ChannelStatus.ON) if turn_on else int(self.ChannelStatus.OFF))))

    def reseta(self):
        self.seta(False)
        return self.seta(True)

    def setb(self, turn_on=True):
        return self.ChannelStatus(int(self.set_value('ec.ChB', int(self.ChannelStatus.ON) if turn_on else int(self.ChannelStatus.OFF))))

    def resetb(self):
        self.setb(False)
        return self.setb(True)

    def get_opt(self):
        return self.CommonOpts( int( self.get_value('ec.optAB', after='now')))

    def set_opt(self, opts):
        assert(type(opts) == self.CommonOpts)
        return self.CommonOpts(int( self.set_value('ec.optAB', int(self.get_opt() | opts))))

    def clr_opt(self, opts):
        assert(type(opts) == self.CommonOpts)
        return self.CommonOpts(int( self.set_value('ec.optAB', int(self.get_opt() & ~opts))))

    def geta_opt(self):
        return self.ChannelOpts( int( self.get_value('ec.optA', after='now')))

    def seta_opt(self, opts):
        assert(type(opts) == self.ChannelOpts)
        return self.ChannelOpts( int( self.set_value('ec.optA', int(self.geta_opt() | opts))))

    def clra_opt(self, opts):
        assert(type(opts) == self.ChannelOpts)
        return self.ChannelOpts(int( self.set_value('ec.optA', int(self.geta_opt() & ~opts))))

    def getb_opt(self):
        return self.ChannelOpts( int( self.get_value('ec.optB', after='now')))

    def setb_opt(self, opts):
        assert(type(opts) == self.ChannelOpts)
        return self.ChannelOpts(int( self.set_value('ec.optB', int(self.getb_opt() | opts))))

    def clrb_opt(self, opts):
        assert(type(opts) == self.ChannelOpts)
        return self.ChannelOpts(int( self.set_value('ec.optB', int(self.getb_opt() & ~opts))))

    def fetch_logs(self, log_range=[1, 10]):

        def decode_logs(logs):
            # find first log with full timestamp, if all is ok it should be the first one
            for log_idx in range(len(logs)):
                header = int(logs[log_idx][4:6], 16)
                if ((header & self.HEADER_MASK) == self.LOG_ID_TIME_CHANGE):
                    break

            if (log_idx != 0):
                _eprint('Timestamp missing, ', log_idx, ' logs discarded.')

            abstime = int(logs[log_idx][12:20], 16)
            time.ctime(abstime)
            start_idx = log_idx

            logs_data = []
            for log_idx in range(start_idx, len(logs), 1):
                header = int(logs[log_idx][4:6], 16) & self.HEADER_MASK

                if ((header == self.LOG_ID_ENERGY) or (header == self.LOG_ID_SUBMETERING) or (header == self.LOG_ID_ENERGY1)):
                    # Energy Log
                    t = int(logs[log_idx][6:10], 16)
                    t_msb = (t & 32768) / 32768
                    t = t * 2
                    t = t & 0xFFFF

                    tlast = abstime & 0xFFFF
                    tlast_msb = (abstime & 65536) / 65536

                    cas = (abstime & 0xFFFF0000)
                    cas = cas + t

                    if (t >= tlast):
                        if (t_msb != tlast_msb):
                            cas = cas + 65536
                    else:
                        if (t_msb != tlast_msb):
                            cas = cas + 65536
                        else:
                            cas = cas + 131072

                    abstime = cas
                    energy = int(logs[log_idx][10:16], 16) * 18e-5
                    tariff = int(logs[log_idx][4:6], 16) & self.HEADER_DATA_MASK
                    energyA = int(logs[log_idx][16:18], 16) / 255 * energy
                    energyB = int(logs[log_idx][18:20], 16) / 255 * energy
                    if (header == self.LOG_ID_ENERGY):
                        notes = 'ECL_SYNC'
                    elif (header == self.LOG_ID_SUBMETERING):
                        notes = 'SUBMETERING'
                    elif (header == self.LOG_ID_ENERGY1):
                        notes = 'TARIFF'
                    else:
                        notes = ''

                    data = dict()
                    data['time'] = abstime
                    data['type'] = 'ENERGY'
                    data['energy'] = energy
                    data['energyA'] = energyA
                    data['energyB'] = energyB
                    data['tariff'] = tariff
                    data['notes'] = notes
                    logs_data.append(data)
                    # print(time.ctime(abstime), '\t', energy, '\t', tariff, '\t ENERGY')

                elif (header == self.LOG_ID_TIME_CHANGE):
                    abstime = int(logs[log_idx][12:20], 16)
                    dT = int(logs[log_idx][6:12], 16)
                    reason = int(logs[log_idx][4:6], 16) & self.HEADER_DATA_MASK
                    if (reason == self.LOG_REASON_SYSTEM_POWER_UP):
                        reason = 'POWER_UP'
                    elif (reason == self.LOG_REASON_WDT_RESET_OCCUR):
                        reason = 'WDT_RESET'
                    elif (reason == self.LOG_REASON_TIME_CHANGE):
                        reason = 'TIME_CHANGE'
                        if (dT == 0):
                            reason = 'WEAK_TIMESTAMP'

                    tariff = int(logs[log_idx][4:6], 16) & self.HEADER_DATA_MASK

                    data = dict()
                    data['time'] = abstime
                    data['type'] = 'TIMECHANGE'
                    data['notes'] = reason
                    data['tariff'] = tariff
                    data['dt'] = dT
                    logs_data.append(data)
                    # print(time.ctime(abstime), '\t', dT, '\t', reason, '\t TIMECHANGE')

                elif (header == self.LOG_ID_FAULT):
                    abstime = int(logs[log_idx][6:14], 16)
                    loadA = int(logs[log_idx][14:16], 16)
                    loadB = int(logs[log_idx][16:18], 16)
                    fault = int(logs[log_idx][18:20], 16)
                    tariff = int(logs[log_idx][4:6], 16) & self.HEADER_DATA_MASK

                    data = dict()
                    data['time'] = abstime
                    data['type'] = 'FAULT'
                    data['tariff'] = tariff
                    data['loadA'] = loadA
                    data['loadB'] = loadB
                    data['fault'] = fault
                    logs_data.append(data)
                    # print(time.ctime(abstime), '\t', loadA, '\t', loadB, '\t', fault, '\t FAULT')

            return logs_data

        logs = []
        search_back_max_idx = max(log_range) + 100
        if (search_back_max_idx > self.EC_LOG_BUF_LEN):
            search_back_max_idx = self.EC_LOG_BUF_LEN

        search_idx = max(log_range)
        logs_buf = []
        timestamp_found = False

        # search for the first available log with complete timestamp
        _eprint("Preparing to fetch logs ...")
        while (search_idx < search_back_max_idx):
            self.set_value('ec.i', search_idx)
            system_args = self.get_args(15)['args']
            logs_buf.append(system_args)

            header = int(system_args[4:6], 16)
            if ((header & self.HEADER_MASK) == self.LOG_ID_TIME_CHANGE):
                timestamp_found = True
                break

            search_idx = search_idx + 1

        if (timestamp_found == False):
            raise ValueError('Log with Timestamp not found!')

            # copy already downloaded logs into buffer, do not download them again
        for i in range(len(logs_buf), 0, -1):
            logs.append(logs_buf[i - 1])

        min_idx = min(log_range) - 1
        if (min_idx < 0):
            min_idx = 0

        for log_idx in range(max(log_range) - 1, min_idx, -1):
            try:
                _eprint("Fetching:", log_idx, " ", end="\r")
                self.set_value('ec.i', log_idx)
                system_args = self.get_args(15)['args']
                logs.append(system_args)
            except:
                _eprint('Error: Stopped due to communication error at log', log_idx)
                return logs

        _eprint("Completed successfully.")
        return decode_logs(logs)

    @staticmethod
    def display_logs(logs_data, log_type='ENERGY'):
        if (log_type == 'ENERGY'):
            print('Time', '\t', 'Log Type', '\t', 'Energy [kWh]', '\t', 'EnergyA [kWh]', '\t', 'EnergyB [kWh]', '\t',
                'Tariff', '\t',
                'Notes')

        for i in range(len(logs_data)):
            if (logs_data[i]['type'] == 'ENERGY' and log_type == 'ENERGY'):
                print(time.ctime(logs_data[i]['time']), '\t', logs_data[i]['type'], '\t',
                    format(logs_data[i]['energy'], '.4f'),
                    '\t', format(logs_data[i]['energyA'], '.4f'), '\t', format(logs_data[i]['energyB'], '.4f'), '\t',
                    logs_data[i]['tariff'], '\t', logs_data[i]['notes'])

            if (logs_data[i]['type'] == 'TIMECHANGE' and log_type == 'TIMECHANGE'):
                print(time.ctime(logs_data[i]['time']), '\t', logs_data[i]['type'], '\t', logs_data[i]['tariff'],
                    '\t', logs_data[i]['dt'], '\t', logs_data[i]['notes'])

            if (logs_data[i]['type'] == 'FAULT' and log_type == 'FAULT'):
                print(time.ctime(logs_data[i]['time']), '\t', logs_data[i]['type'], '\t', logs_data[i]['tariff'],
                    '\t', logs_data[i]['loadA'], '\t', logs_data[i]['loadB'], '\t', logs_data[i]['fault'])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent('''\
            Energy Control Modules Utility
            ----------------------------------------------------------------------
            Provide the device name of which module's you wish to fetch
            logs with -d option or leave blank to show available EC modules.
           '''))
    parser.add_argument("--gateway", help="gateway address, defaults to http://localhost:33000", required=False,
                        default='http://localhost:33000')
    parser.add_argument("--username", help="gateway username", required=False)    
    parser.add_argument("--password", help="gateway password", required=False)    
    parser.add_argument("-d",
                        help="Device to dump log data, like: 'EC-D16-20AHR-L2A-1.2.0-1901-1 50678419 (00043E29B4CD)'",
                        required=False, default=None)
    parser.add_argument("-N", help="Number of logs to dump, default is 20 and max is 992", required=False, type=int,
                        default=20)
    parser.add_argument("-t", help="Type of log, defaults to ENERGY", required=False, default='ENERGY',
                        choices=['ENERGY', 'FAULT', 'TIMECHANGE'])
    args = parser.parse_args()

    gw = gateway.Group(gateway=args.gateway, username=args.username, password=args.password)
    last_log = args.N
    if last_log > EC.EC_LOG_BUF_LEN:
        last_log = EC.EC_LOG_BUF_LEN
        _eprint("EC may contains up to", EC.EC_LOG_BUF_LEN, "logs.")

    if args.d:
        ec = EC(gw, args.d)
        logs = ec.fetch_logs([1, last_log])
        EC.display_logs(logs, args.t)
    else:
        for x in gw.get_device_list():
            if 'EC-' in x['name']:
                _eprint(x['root'], x['name'], EC(gw, x['name']).get_status() )



"""ISOTEL IDM Support for Monodaq U Series
   http://isotel.eu/idm/device/monodaq/monodaq-u-x.html
"""

from isotel.idm import gateway, signal, _version
import socket
import struct
import time
import requests
import logging
import argparse
try:
    if get_ipython:
        from IPython.core.display import display, HTML
except:
    pass

class MonoDAQError(Exception):
    def __init__(self, message):
        super().__init__(message)

class MonoDAQ_U(gateway.Device):
    USER_HEADER_LEN = 2
    DIGITAL_MASK = 0x80080000   # digital stream enabled with additional bit in digital only mode
    ANALOG_MASK = 0x47FFFF      # all analog channels present in a stream
    ANUSED_MASK = 0x7FFFF       # used (selected) channels
    PCKTCNT_MASK = 0x3FFF       # counter mask to detect lost packets

    """
    MonoDAQ_U Class adds triggering, streaming data retrieval methods and plots
    """

    def __init__(self, group = None, name = None, exclude = None, accuracy=True, prefix=''):
        """
        :param group: an gateway.Group reference
        :param str name: name of the device returned by the get_device_list(),
                        such as 'device0' root designation or a full name
        """
        super().__init__(group, name, exclude, unit=True, accuracy=accuracy, advanced=True, development=True, hidden=True, brand_filter='MonoDAQ-U-')
        self.analog_channels_map = None
        self.digital_channels_map = None
        self.ad_mux_copy = None
        self.analog_socket = None
        self.digital_socket = None
        self.ad_mux = None
        self.channel_muxes = None
        self.units = None
        self.pins = None
        self.ch_gains = None
        self.ain_sa_pckt = None
        self.din_sa_pckt = None
        self.din2ain_rate = None
        self.dt_ain = None
        self.dt_din = None
        self.fetched_Ain = None
        self.fetched_Din = None
        self.pckt_cnt_Ain = 0
        self.pckt_cnt_Din = 0
        self.ain = b''
        self.din = b''
        self.ain_last_T = 0
        self.din_last_T = 0
        self.prefix = prefix.upper()        
        self.settime(False)
        self.logger = logging.getLogger(__name__)

    def reset(self):
        self['daq.trigger.timed'] = 'Off'  # Stop A/D
        self['daq.trigger.N']     = '0'
        if self['ch.config.state'] != 'Cleared':
            self['ch.config.state']   = '~Clear' # Reset channel setup configuration
        self.get('ch', after='now') # Update entire table
    
    def settime(self, finetune=False):
        if self.get_value('id.b1') == 'Uninitialized' or not self.active:
            return None

        t = self.get('clock', after='now')
        t_diff = float(t['time'])*1000 - float(t['clock']['time']['value'])

        # Sync time to UTC, always when not initialized otherwise on request
        if finetune or abs(t_diff) > 321408000000:
            t['clock']['time']['value']  = '~' + str((round(t_diff)))
            t['clock']['uticks']['value']  = '~0'
            t.pop('status')
            t.pop('time')
            self.set('clock', t)

        return -t_diff # [ms]

    def isready(self):
        sys_state = self.get_value('sys.state', after='now')
        return sys_state == 'Ready' or sys_state == 'Uncalibrated'

    def wait4ready(self, mintime=0, timeout=15):
        mintime = timeout - mintime
        while not self.isready() or timeout > mintime:
            time.sleep(0.25)
            timeout -= 0.25
            if timeout <= 0:
                raise MonoDAQError('Device is not ready because it is: ' + self.get_value('sys.state'))

    def wait4value(self, ch_num):
        while not (self['ch.value.value' + str(ch_num)] != '---'):
            pass

    def get_streaming_channels(self):
        return [int(''.join(list(filter(str.isdigit, ch[0])))) for ch in self.get('ch.mux')['mux'].items() if int(ch[1]['value'],16)]

    def is_digital(self, channel):
        return int(self['ch.mux.mux' + str(channel)],16) & self.DIGITAL_MASK

    def get_channel_bypin(self, pin_name):
        return [int(''.join(list(filter(str.isdigit, x[0])))) for x in self.get('ch.pin')['pin'].items() if x[1]['value'] == pin_name][0]
    
    def get_setup(self, HTML=None):
        """
        :returns: html description of the channel setup
        """
        if HTML:
            resp = self.request.get(self.get_url() + '/ch.html', self.group.timeout)
        else:
            resp = self.request.get(self.get_url() + '/ch.txt', self.group.timeout)
            
        return resp

    def set_channel_value(self, pin_name, value, after=None):
        channel = self.get_channel_bypin(pin_name)
        return self.set_value('ch.set.set'+str(channel), value, after=after)

    def get_channel_value(self, pin_name, after=None):
        channel = self.get_channel_bypin(pin_name)
        return self.get_value('ch.set.set'+str(channel), after=after)

    def print_setup(self):
        try:
            get_ipython
            display(HTML(self.get_setup(HTML=True)))
        except:
            print(self.get_setup(HTML=False))

    def stop(self):
        self['daq.trigger.timed'] = 'Off'
        self.ain = b''
        self.din = b''

    def _fetch_init(self, N=0, timed=None):

        def get_gains(ad_mux, ch_muxes, ch_value_config):
            gain = [[]] * bin(ad_mux & self.ANALOG_MASK).count('1')
            for ch_num in range(len(ch_muxes)):
                ch_mux = int(ch_muxes['mux' + str(ch_num)]['value'], 16)
                if ch_mux & self.ANALOG_MASK:
                    ch_idx = bin(ad_mux & int(ch_mux * 2 - 1)).count('1') - 1
                    gain[ch_idx] = 1/float(ch_value_config['value']['value'+str(ch_num)]['args'][-1])
            return gain

        def port_tcp(port):
            return int(self.get_protocols()['LongTransport'][str(port)]['SFrame']
                       ['Stream' + str(port)]['status']['port'])

        # Fetch channel configuration
        self.wait4ready()
        daqconf = self.get('daq')
        self.ad_mux = int(daqconf['daq']['ad']['mux']['value'], 16)
        self.channel_muxes = self.get('ch.mux')['mux']
        self.units = self.get('ch.unit')['unit']
        self.pins  = self.get('ch.pin')['pin']
        self.ch_gains = get_gains(self.ad_mux, self.channel_muxes, self.get('ch.value', args=True))

        # Close possibly open streams if mux has changed
        if self.ad_mux_copy:
            #if self.ad_mux != self.ad_mux_copy:
            if self.analog_socket:
                self.analog_socket.close()
                self.analog_socket = None
            if self.digital_socket:
                self.digital_socket.close()
                self.digital_socket = None

        self.ad_mux_copy = self.ad_mux

        # Open TCP/IP Streams
        # digital input only case
        if (self.ad_mux & self.DIGITAL_MASK) == self.ad_mux:
            if self.analog_socket:
                self.analog_socket.close()
                self.analog_socket = None
        else:
            if self.ad_mux & self.ANALOG_MASK and not self.analog_socket:
                self.analog_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.analog_socket.connect((self.group.get_gatewayname(), port_tcp(1)))
                self.analog_socket.settimeout(0.1)

        if self.ad_mux & self.DIGITAL_MASK and not self.digital_socket:
            self.digital_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.digital_socket.connect((self.group.get_gatewayname(), port_tcp(2)))
            self.digital_socket.settimeout(0.1)

        if self.analog_socket == None and self.digital_socket == None:
            raise MonoDAQError('No streaming channels selected')

        # Metrics used in decoding and validation
        self.ain_sa_pckt = int(daqconf['daq']['ad']['AI']['value']) if self.analog_socket else 0
        self.din_sa_pckt = int(daqconf['daq']['ad']['DI']['value'])
        self.din2ain_rate = int(daqconf['daq']['ad']['DI_AI']['value'])
        self.dt_ain      = 1 / float(daqconf['daq']['ad']['rate']['value'])
        self.dt_din      = self.dt_ain / self.din2ain_rate if self.din2ain_rate > 0 else 0
        self.fetched_Ain = 0
        self.fetched_Din = 0
        self.pckt_cnt_Ain= 0
        self.pckt_cnt_Din= 0
        self.ain_last_T  = 0
        self.din_last_T  = 0
        self.ain = b''
        self.din = b''
        self.set('daq.trigger',
            dict({'trigger': {'timed': {'value': timed if timed else '~Immediate'}, 'N': {'value': N}}}))

    def fetch(self, N=10, timed=None):

        def idm_get_rxcnts(ref):
            stats = self.get_protocols()
            return (stats['LongTransport']['1']['SFrame']['Stream1']['status']['rx'] - ref[0],
                    stats['LongTransport']['2']['SFrame']['Stream2']['status']['rx'] - ref[1])

        # Read data via tcp/ip channels and return when minimum required amount is obtained
        def readdata():
            RETRY_COUNT = 50
            BUF_LEN = 65536*8
            nonlocal ain_recv, din_recv, idm_cnts
            retry = RETRY_COUNT
            while retry > 0:
                try:
                    if ain_recv > ain_len or din_recv > din_len:
                        raise MonoDAQError("Received excess data: analog %d/%d B in buf %d, digital %d/%d B in buf %d" % 
                                           (ain_recv, ain_len, len(self.ain), din_recv, din_len, len(self.din) ))

                    if ((ain_recv >= ain_len and din_recv >= din_len) or
                        (ain_recv >= ain_len and len(self.din) >= din_lenmin) or
                        (din_recv >= din_len and len(self.ain) >= ain_lenmin) or                        
                        (len(self.din) >= din_lenmin and len(self.ain) >= ain_lenmin)):     # Local condition
                        break

                    if self.analog_socket and ain_recv < ain_len:
                        d = self.analog_socket.recv(BUF_LEN)
                        self.ain += d
                        ain_recv += len(d)
                    if self.digital_socket and din_recv < din_len:
                        d = self.digital_socket.recv(BUF_LEN)
                        self.din += d
                        din_recv += len(d)
                    
                    retry = RETRY_COUNT
                except socket.timeout:
                    self.logger.debug("sample_ad: socket.timeout, din { req:%d, len:%d }, ain: { req:%d, len:%d }",
                                      din_len, din_recv, ain_len, ain_recv)
                    retry -= 1
                    if retry == 0:
                        idm_resent = idm_get_rxcnts(idm_cnts)
                        raise MonoDAQError("Did not receive enough data: analog %d/%d B in buf %d, digital %d/%d B in buf %d, idm resent analog: %d B, digital: %d B" % 
                                           (ain_recv, ain_len, len(self.ain), din_recv, din_len, len(self.din), idm_resent[0], idm_resent[1] ))
                    pass

        def decode_analog(samples_ppck, dt, demuxchs, gains, ain, t_offset):
            T = 0
            x = []
            y = [[]]*anused_chs
            if anused_chs == 0:
                return x,y

            packet_size = 2*samples_ppck            
            assert( int(samples_ppck/demuxchs) == float(samples_ppck/demuxchs) )
            for frame in range(self.USER_HEADER_LEN, len(ain), int(packet_size)+self.USER_HEADER_LEN):
                if frame+packet_size <= len(ain):

                    # Detect possible losses on the way
                    if self.pckt_cnt_Ain != (((ain[frame-self.USER_HEADER_LEN+1]<<8) + ain[frame-self.USER_HEADER_LEN]) & self.PCKTCNT_MASK):
                        raise MonoDAQError('Detected lost analog streaming packets')
                    
                    self.pckt_cnt_Ain = (self.pckt_cnt_Ain + 1) & self.PCKTCNT_MASK

                    decoded = struct.unpack(str(samples_ppck) + "h", ain[frame:frame+packet_size])
                    for ch in range(anused_chs):
                        if not y[ch]:
                            y[ch] = []
                            
                        y[ch] += [ int(y*gains[ch]*1000000+0.5)/1000000 for y in decoded[ch::demuxchs] ]
                    x += [int(((t_offset + t*dt)*1000000)+0.5)/1000000 for t in range(T,T+int(samples_ppck/demuxchs))]
                    T += int(samples_ppck/demuxchs)
                else:
                    print("Analog Stream has received inconsistent data", len(ain), 'required', frame+packet_size+self.USER_HEADER_LEN)
            return x,y
            
        def getanalog():
            Ax,Ay = decode_analog(self.ain_sa_pckt, self.dt_ain, ain_chs, self.ch_gains, self.ain[:ain_lenmin], self.fetched_Ain * self.dt_ain)
            self.fetched_Ain += len(Ax)
            self.ain = self.ain[ain_lenmin:]
            return Ax,Ay

        def getanalog_labels(demuxchs, pinname=True):
            labels = [None]*demuxchs
            for ch_num in range(len(self.channel_muxes)):
                ch_mux = int(self.channel_muxes['mux' + str(ch_num)]['value'], 16)
                if ch_mux & self.ANALOG_MASK:
                    ch_idx = bin(self.ad_mux & int(ch_mux * 2 - 1)).count('1') - 1
                    if pinname:
                        labels[ch_idx] = self.prefix + self.pins['pin'+str(ch_num)]['value'] + " [" + self.units['unit'+str(ch_num)]['value'] + "]"
                    else:
                        labels[ch_idx] = "stream" + str(ch_num)
                    
            return labels

        def decode_digital(samples_ppck, dt, din, t_offset, dmux, expand=True):
            T = 0
            x = []
            y = []
            for frame in range(self.USER_HEADER_LEN, len(din), int(samples_ppck)+self.USER_HEADER_LEN):
                if frame+samples_ppck <= len(din):

                    # Detect possible losses on the way
                    if self.pckt_cnt_Din != (((din[frame-self.USER_HEADER_LEN+1]<<8) + din[frame-self.USER_HEADER_LEN]) & self.PCKTCNT_MASK):
                        raise MonoDAQError('Detected lost digital streaming packets')
                    
                    self.pckt_cnt_Din = (self.pckt_cnt_Din + 1) & self.PCKTCNT_MASK

                    decoded = struct.unpack(str(samples_ppck) + "B", din[frame:frame+samples_ppck])
                    if expand:                        
                        for z in decoded:
                            Y = [(z>>v) & 1 for i,v in enumerate(dmux)]
                            y += [tuple(Y)]
                    else:
                        y += decoded

                    x += [int(((t_offset + t*dt)*1000000)+0.5)/1000000 for t in range(T,T+samples_ppck)]
                    T += samples_ppck
                else:
                    self.logger.error("Digital Stream has received inconsistent data len:%d, required:%d"
                                      % (len(din), frame + samples_ppck + self.USER_HEADER_LEN))
            return x,y

        def getdigital(dmux):
            Dx,Dy = decode_digital(self.din_sa_pckt, self.dt_din, self.din[:din_lenmin], self.fetched_Din * self.dt_din, dmux)
            self.fetched_Din += len(Dx)
            self.din = self.din[din_lenmin:]
            return Dx,Dy

        def getdigital_labels(pinname=True):
            labels = []
            dmux   = []
            for ch_num in range(len(self.channel_muxes)):
                if int(self.channel_muxes['mux' + str(ch_num)]['value'], 16) & self.DIGITAL_MASK:
                    assert(ch_num >= 0 and ch_num <= 7)
                    dmux += [ch_num]
                    if pinname:
                        labels += [ self.prefix + self.pins['pin'+str(ch_num)]['value'] + " [" + self.units['unit'+str(ch_num)]['value'] + "]" ]
                    else:
                        labels += [ "din" + str(ch_num) ]
            return labels, dmux

        # Prepare HW for fetch
        self._fetch_init(N, timed)

        # Calculate total number of packets to be expected and minimal lengths to be
        # able to decode a frame
        ain_len   = 0
        ain_pckts = 0
        anused_chs= bin(self.ad_mux & self.ANUSED_MASK).count('1')
        ain_chs   = bin(self.ad_mux & self.ANALOG_MASK).count('1')
        ain_recv  = 0
        ain_lenmin= 0
        if self.ain_sa_pckt > 0:
            ain_pckts = int( (N*ain_chs + self.ain_sa_pckt-1) / self.ain_sa_pckt )
            ain_len   = ain_pckts * (2*self.ain_sa_pckt + self.USER_HEADER_LEN)
            ain_lenmin= 2*self.ain_sa_pckt + self.USER_HEADER_LEN

        din_len   = 0
        din_pckts = 0
        din_recv  = 0
        N_digital = 0
        din_lenmin= 0
        if self.din_sa_pckt:
            N_digital = self.din2ain_rate * N
            din_pckts = int( (N_digital + self.din_sa_pckt-1) / self.din_sa_pckt )
            din_len   = din_pckts * (self.din_sa_pckt + self.USER_HEADER_LEN)
            din_lenmin= self.din_sa_pckt + self.USER_HEADER_LEN

        #print('total requirements:', N, ain_pckts, ain_len, len(self.ain), ':', N_digital, din_pckts, din_len, len(self.din))
        idm_cnts = idm_get_rxcnts( (0,0) )
        dlabels, dmux = getdigital_labels()
        analog_nans  = tuple( [float('nan')]*anused_chs )
        digital_nans = tuple( [float('nan')]*len(dmux) )
        try:
            yield ("t [s]", tuple(getanalog_labels(anused_chs))) , ("t [s]", tuple(dlabels))
            readdata()
            Ax,Ay = getanalog()
            Dx,Dy = getdigital(dmux)
            Axi = iter(Ax)
            Ayi = iter(zip(*Ay))
            Dxi = iter(Dx)
            Dyi = iter(Dy)
            ax = next(Axi, None)
            ay = next(Ayi, None)
            dx = next(Dxi, None)
            dy = next(Dyi, None)
            i  = 0
            while ax != None or dx != None:
                if ax != None and ax == dx:
                    yield (ax,ay),(dx,dy)
                    ax = next(Axi, None)
                    ay = next(Ayi, None)
                    dx = next(Dxi, None)
                    dy = next(Dyi, None)
                    i += 1
                elif dx == None or (ax != None and ax < dx):
                    yield (ax,ay),(None, digital_nans)
                    ax = next(Axi, None)
                    ay = next(Ayi, None)
                    i += 1
                elif ax == None or (dx != None and ax > dx):
                    yield (None, analog_nans),(dx,dy)
                    dx = next(Dxi, None)
                    dy = next(Dyi, None)
                    if Ax == []:
                        i += 1

                if i >= N:
                    break

                if (Ax != None and ax == None) or (Dx != None and dx == None):
                    assert(i < N)
                    readdata()

                if Ax != [] and ax == None:
                    Ax,Ay = getanalog()
                    Axi = iter(Ax)
                    Ayi = iter(zip(*Ay))
                    ax = next(Axi, None)
                    ay = next(Ayi, None)

                if Dx != [] and dx == None:
                    Dx,Dy = getdigital(dmux)
                    Dxi = iter(Dx)
                    Dyi = iter(Dy)
                    dx = next(Dxi, None)
                    dy = next(Dyi, None)

        except KeyboardInterrupt:
            self.stop()

        except (GeneratorExit, MonoDAQError):
            self.stop()
            raise

        self.stop()


class MonoDAQs_U(object):
    def __init__(self, *monodaqs):
        self.group = None
        self.monodaqs = monodaqs
        # TODO 1-wire config and auto config prefixes

    def fetch(self, N):
        for d in self.monodaqs:
            d.wait4ready()

        # TODO detect false starts
        t_start = float(self.monodaqs[0].get('clock', after='now')['clock']['time']['value']) + 2000 # ms
        streams = [[]]*len(self.monodaqs)
        for i,d in enumerate(self.monodaqs):
            streams[i] = d.fetch(N, t_start)
        return signal.mergeNsync( *streams )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='MonoDAQ-U-X Utility %s' % (_version.__version__))

    parser.add_argument("--gateway", help="gateway address, defaults to http://localhost:33000", required=False,
                        default='http://localhost:33000')
    parser.add_argument("--user", help="user name", required=False, default=None)
    parser.add_argument("--pass", dest='password', help="password", required=False, default=None)
    parser.add_argument("-d",
                        help="Device to fetch from, like: 'MonoDAQ-U-X v1.55.0.17 0x4' or just: device0",
                        required=False)
    parser.add_argument("-N", help="Fetch N samples and print results in TSV format", 
                        required=False, type=int)
    parser.add_argument("--from", dest='start', help="Fetch samples from records, given by time in UTC, or bounded in interval --to", required=False, default=None)
    parser.add_argument("--to", dest='end', help="Fetch samples from records until time in UTC, or bounded in interval --from", required=False, default='last')
    parser.add_argument("--scale", help="Down-sample samples from records to given number", required=False, default=None)
    parser.add_argument("-s", help="Set or query a parameter, like: ch.function.function0 'Digital Input'", required=False, nargs='+')

    args = parser.parse_args()

    try:
        gw = gateway.Group(gateway=args.gateway, username=args.user, password=args.password)
    except:
        print('Cannot connect to a gateway %s, use --gateway http://yourgw:33000 [--user USERNAME] [--pass PASSWORD]' % ( args.gateway ) )
        exit(-1)
    
    if args.d:
        mdu = MonoDAQ_U(gw, name = args.d)
        if args.start or args.end:
            if args.s:
                records = mdu.get_records(args.s[0], time_from=args.start, time_to=args.end, limit=args.N, scale=args.scale)
                for item in records:
                    try:
                        print(item['time'], end='\t')
                        if item['value']:
                            for k,v in item['value'].items():
                                try:
                                    print(v['value'], end='\t')
                                except:
                                    pass
                    except:
                        for k,v in item.items():
                            try:
                                print(v['value'], end='\t')
                            except:
                                pass                                
                    print('')
            else:
                print('Specify a parameter to dump')
        elif args.s:
            if len(args.s) == 2:
                mdu[args.s[0]] = args.s[1]
            if len(args.s) >= 1:
                print(mdu.get_value(args.s[0], after='now'))
        elif args.N:
            for line in signal.to_tsv( mdu.fetch(args.N) ):
                print(line)
        else:
            mdu.print_setup()
    else:
        for d in gw.get_device_list(brand_filter='MonoDAQ-U'):
            print(d['root'], '\t', d['name'])





"""ISOTEL Data-acqusition & Metering (IDM) Smart Gateway API
"""

from __future__ import absolute_import
import json
from uncertainties import ufloat, ufloat_fromstr

try:
    from . import _urllib3 as gateway_urllib
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
    from . import _urllib2 as gateway_urllib


def get_compatible_revision():
    return 845

def get_compatible_major_revision():
    return "1.2"

def check_compatible_version(gateVer, compVer, majorVer):
    """
    Check version compatibility
    
    :param sver: 
    :param lver: 
    :return: 
    """
    #print(str(sver) + " vs: " + str(compVer))
    if not str(gateVer).startswith(majorVer):
        raise RuntimeError('Major API versions not compatible, gateway: {} library: {}'.format(gateVer[:3], majorVer))

    rev = int(str(gateVer).split('.')[-1:][0])
    if rev != compVer  :
        raise RuntimeError('API versions not compatible, gateway: {} library: {}, {} upgrade required.'
                           .format(str(rev), str(compVer), "gateway" if rev < compVer else "library"))


def is_compatible_legacy_version(sver):
    return int(sver) <= 882


class GatewayError(Exception):
    def __init__(self, message):
        super().__init__(message)


class Group:
    """Attach to a Group instance of an IDM Gateway,
    which group can spread in a cluster over several IDM Gateways.

    :param str gateway: gateway url, such as 'http://localhost:33000'
    :param str group: a group of devices that belong to one entity,
                      as 'me' to access local devices or 'friend1'
    :param int timeout: request timeout in seconds
    :param bool ssl_untrusted: request lets true untrusted ssl certificate
    :param str ssl_cert_file: client ssl certificate file
    :param str ssl_cert_pass: client ssl certificate password
    """

    def __init__(self, gateway='http://localhost:33000', group='me', timeout=60, ssl_untrusted=True, ssl_cert_file=None,
                 ssl_cert_pass=None, username=None, password=None, keep_alive=True, keep_alive_timeout=10):

        version = 'v1'
        self.params = {
            "gateway": gateway,
            "group": group,
            "timeout": timeout,
            "version": version,
            "ssl_untrusted": ssl_untrusted,
            "ssl_cert_file": ssl_cert_file,
            "ssl_cert_pass": ssl_cert_pass,
            "username": username,
            "password": password,
            "keep_alive": keep_alive,
            "keep_alive_timeout": keep_alive_timeout

        }
        self.gateway = gateway + '/api/' + version
        self.group = group
        self.gateway_url = self.gateway + '/' + group
        self.timeout = timeout
        self.context = None
        self.auth_header = None

        if username and password:
            self.auth_header = gateway_urllib.encode_auth_header(username, password)

        if (self.gateway_url.startswith('https://')):
            self.context = gateway_urllib.get_context(untrusted=ssl_untrusted, cert_file=ssl_cert_file,
                                                      cert_pass=ssl_cert_pass)

        self.request = gateway_urllib.RequestHandler(self.context, self.auth_header,
                                                     keep_alive=keep_alive, keep_alive_timeout=keep_alive_timeout)

        sver = self.request.get(self.gateway_url + '/version.json', self.timeout)
        check_compatible_version(sver['api'], get_compatible_revision(), get_compatible_major_revision())

    def clone(self):
        """
        Clone this Group
        :return: 
        """
        me = Group(self.params['gateway'],
                   self.params['group'],
                   timeout=self.params['timeout'],
                   ssl_untrusted=self.params['ssl_untrusted'],
                   ssl_cert_file=self.params['ssl_cert_file'],
                   ssl_cert_pass=self.params['ssl_cert_pass'],
                   username=self.params['username'],
                   password=self.params['password'],
                   keep_alive=self.params['keep_alive'],
                   keep_alive_timeout=self.params['keep_alive_timeout'])
        return me

    def get_gatewayname(self):
        """
        :returns: gateway name
        """
        return urlparse(self.gateway).netloc.split(':')[0]

    def get(self, uri, params=None):
        """
        Sends generic GET request
        
        :param str uri: resource uri
        :param dict params: HTTP request parameters
        :returns: Resource response
        :rtype: depending on resource
        :raises: HTTPError, socket.timeout
        """
        if not uri.startswith("/"):
            uri = "/" + uri
        uri = self.gateway + uri
        uri = self.add_params_to_uri(uri, params=params)

        return self.request.get(uri, self.timeout)

    def post(self, uri, data, dump_json=True):
        """
        Sends generic POST request
        
        :param str uri: resource uri
        :param dict data: Request body
        :param dump_json: convert json data to string if True
        :returns: Resource response
        :rtype: depending on resource
        :raises: HTTPError, socket.timeout
        """
        if not uri.startswith("/"):
            uri = "/" + uri
        uri = self.gateway + uri
        if dump_json:
            data = json.dumps(data)
        return self.request.post(uri, data, self.timeout)

    def delete(self, uri, params=None):
        """
        Sends generic DELETE request
        
        :param str uri: resource uri
        :param dict params: HTTP request parameters
        :returns: Resource response
        :rtype: depending on resource
        :raises: HTTPError, socket.timeout
        """
        if not uri.startswith("/"):
            uri = "/" + uri
        uri = self.gateway + uri
        uri = self.add_params_to_uri(uri, params=params)

        return self.request.delete(uri, self.timeout)

    def get_device_list(self, brand_filter=None, inactive=False):
        """Gets list of devices from gateway.

        :returns: list of devices
        :rtype: list of dicts
        :raises: HTTPError, socket.timeout
        """
        uri = self.gateway_url + ".json"
        if inactive==False:
            uri += "?active=true"
        
        devlist = self.request.get(uri, self.timeout)
        if brand_filter:
            devlist = [i for i in devlist if brand_filter in i['name'] ]

        return devlist

    def get_service_list(self):
        """Gets list of running services from gateway.

        :returns: list of devices
        :rtype: list of dicts
        :raises: HTTPError, socket.timeout
        """
        uri = self.gateway_url + "/services.json"

        return self.request.get(uri, self.timeout)

    def get_service_pin(self, name):
        """Request service PIN number required for connection
        :returns: pin number
        """
        uri = self.gateway_url + "/services/pin/" + str(name)

        return self.request.get(uri, self.timeout)

    def get_bluetooth_list(self, after=None):
        """Fetches list of devices connected through bluetooth.

        :returns: list of bluetooth devices
        :rtype: list of dicts
        :raises: HTTPError, socket.timeout
        """
        return self._get_list(self.gateway_url + "/bluetooth.json", after=after, timeout=90)

    def get_serial_list(self, after=None):
        """Fetches list of devices connected through serial ports.

        :returns: list of serial devices
        :rtype: list of dicts
        :raises: HTTPError, socket.timeout
        """
        return self._get_list(self.gateway_url + "/serial.json", after=after, timeout=30)

    def _get_list(self, uri, after=None, timeout=None):
        if after:
            uri += '?after=' + str(after)
        return self.request.get(uri, timeout if timeout else self.timeout)

    def get_connections(self):
        """
        Get connections
        
        :returns: adapter status in json form
        """
        uri = self.gateway_url + "/connections"

        return self.request.get(uri, self.timeout)

    def get_protocols(self, port):
        """
        Get protocol stack dump for given adapter/port

        :returns: adapter status in json form
        """
        uri = self.gateway_url + "/connections/" + port

        return self.request.get(uri, self.timeout)

    def get_adapter_status(self, adapter, level=None, connection='serial'):
        """
        Get connected adapter status
        
        :param str adapter: adapter preset or communication port
        :param str level: limit status results by protocol level (phy, frame, ec, trans, msg)
        :param str connection: can be (serial)
        :returns: adapter status in json form
        """
        uri = self.gateway_url + "/" + connection + "/" + adapter
        if level:
            uri += "/" + level
        return self.request.get(uri, self.timeout)

    def set_adapter_setting(self, adapter, level, data):
        """
        Set protocol parameter
        
        :param str adapter: adapter preset or communication port
        :param str level: limit status results by protocol level (phy, frame, ec, trans, msg)
        :param data: data format {"setting_name":"value"}
        :returns: operation result
        """
        uri = self.gateway_url + "/serial/" + adapter + "/" + level
        return self.request.post(uri, json.dumps(data), self.timeout)

    def get_activity(self, params=None):
        """
        Get latest gateway activity logs
        
        :param int limit: max number of returned logs
        :returns: activity logs
        :rtype: list of dicts
        :raises: HTTPError, socket.timeout
        """
        uri = self.gateway_url + '/activity.json'
        uri = self.add_params_to_uri(uri, params=params)

        return self.request.get(uri, self.timeout)

    def push_notification(self, data):
        """
        Sends generic POST request
        
        :param str uri: resource uri
        :param dict data: Request body
        :param dump_json: convert json data to string if True
        :returns: Resource response
        :rtype: depending on resource
        :raises: HTTPError, socket.timeout
        """
        uri = self.gateway_url + '/activity.json'
        return self.request.post(uri, json.dumps(data), self.timeout)

    def time(self):
        """Returns current gateway time.

        :returns: current gateway time
        :rtype: float
        """
        ts = self.request.get(self.gateway_url + '/servertime.json', self.timeout)
        try:
            return float(ts['time'])
        except:
            return None

    def version(self):
        """Returns current gateway time.

        :returns: current gateway time
        :rtype: float
        """
        return self.request.get(self.gateway_url + '/version.json', self.timeout)

    def run_script(self, script_name, phy_name):
        """Runs a script on a gateway.

        :returns: state of the script.
        """
        data = {'alias': script_name, 'phy': phy_name}
        return self.request.post(self.gateway + "/terminal/aliases",
                                 json.dumps(data), self.timeout)

    def kill_script(self):
        """Kills currently running script on a gateway.

        :returns: state of the last script
        """
        self.do('$kill')

    def do(self, command, params=''):
        """Sends a command to a gateway.
        Refer to gateway help for a list of available commands, obtainable with the $help command.

        :param str command: name of the command to be executed
        :param str params: optional parameters for the command
        """
        data = {
            'command': command,
            'parameters': params
        }
        return self.request.post(self.gateway + "/terminal/commands",
                                 json.dumps(data), self.timeout)

    def load_virtual_device(self, data, owner=None):
        """
        Create virtual device
        """
        uri = self.gateway_url + "/virtual"
        return self.request.post(uri, json.dumps(data), self.timeout)

    def remove_device(self, device):
        """
        Remove given active device from protocol stack
        """
        uri = self.gateway_url + "/" + str(device)
        return self.request.delete(uri, self.timeout)

    def get_property_list(self, identifier):
        """Retrieves a list of properties related to the device.
        :returns: list of property names
        :rtype: list
        """
        uri = self.gateway_url + "/"
        uri += identifier + "/properties" + '.json'
        return self.request.get(uri, self.timeout)

    def add_params_to_uri(self, uri, params=None):
        if params:
            uri += '?' + '&'.join({"%s=%s" % (k, v) for (k, v) in params.items()})
        return uri

    def get_property(self, identifier, prop, params=None):
        """Fetches value of *prop* property.
        The object returned has a definite structure:
        ``{ prop: {data: <value>}, time: <tstmp> }``

        :param str identifier: property group identifier
        :param str prop: name of the property
        :returns: value of the property
        :rtype: dict
        """
        prop = prop.replace(".", "/")
        uri = self.gateway_url + "/"
        uri += identifier + "/properties/" + prop + '.json'

        uri = self.add_params_to_uri(uri, params=params)

        return self.request.get(uri, self.timeout)

    def set_property(self, identifier, prop, value, append=False):
        """Sets or overwrites the value of the *prop* property with *val* value.
        The *val* passed is stored and enclosed in:
        ``{prop: {data: <val>}}``

        :param str identifier: property group identifier
        :param str prop: name
        :param str|dict|number value: value
        :returns: result with structure ``{result: Error|OK}``
        :rtype: dict
        """
        prop = prop.replace(".", "/")
        uri = self.gateway_url + "/"
        uri += identifier + "/properties/" + prop + '.json'

        if append:
            uri += '?append=true'

        return self.request.post(uri, json.dumps(value), self.timeout)

    def delete_property(self, identifier, prop, params=None):
        """Deletes the `prop` property.
        The entire property, not only its value, is removed.
        
        :param str identifier: property group identifier
        :param str prop: name
        :return: response with structure: ``{result: Error|OK}``
        :rtype: dict
        """
        prop = prop.replace(".", "/")
        uri = self.gateway_url + "/"
        uri += identifier + "/properties/" + prop + '.json'
        uri = self.add_params_to_uri(uri, params=params)

        return self.request.delete(uri, self.timeout)

    def get_arptable(self, params=None):
        """
        Get arptable
        :param params: 
        :return: 
        """

        uri = self.gateway + "/arptable"
        uri = self.add_params_to_uri(uri, params=params)

        return self.request.get(uri, self.timeout)

    def credentials(self):
        """
        Get current user credentials
        """
        uri = self.gateway + "/credentials"

        return self.request.get(uri, self.timeout)

    def get_users(self, principal_filter=None, username=None, key=None, pending=False, user_type=None):
        """
        Get user credentials according to given filters
        
        :param str filter: user principal info filter
        :param str key: user public key
        :param bool pending: pending users
        :param cred_type: credentials type, either password or certificate
        """
        uri = self.gateway + "/credentials/users"
        if pending:
            uri = uri + "/pin"
        if key:
            uri += "?key=" + str(key)
        elif principal_filter:
            uri += "?filter=" + str(principal_filter)
        elif username:
            uri += "?username=" + str(username)

        if user_type:
            uri = uri + ("&" if (key or principal_filter or username) else "?") + "type=" + str(user_type)

        return self.request.get(uri, self.timeout)

    def set_user(self, value, pending=False):
        """
        Set user credentials 
        
        :param str value: credentials json data
        
        """
        uri = self.gateway + "/credentials/users"
        if pending:
            uri = uri + "/pin"
        return self.request.post(uri, json.dumps(value), self.timeout)

    def delete_user(self, principal_filter=None, username=None, key=None, pending=False, user_type=None):
        """
        Delete user credentials
        
        :param str filter: user principal info filter
        :param str key: user public key     
        :param bool pending: pending pin users         
        :param cred_type: credentials type, either password or certificate
        """
        uri = self.gateway + "/credentials/users"
        if pending:
            uri = uri + "/pin"

        if key:
            uri += "?key=" + str(key)
        elif principal_filter:
            uri += "?filter=" + str(principal_filter)
        elif username:
            uri += "?username=" + str(username)

        if user_type:
            uri = uri + ("&" if (key or principal_filter or username) else "?") + "type=" + str(user_type)

        return self.request.delete(uri, self.timeout)

    def register_user(self, pin, name=None):
        """
        Register user credentials with PIN number 
        
        :param int pin: PIN number
        
        """
        uri = self.gateway + "/credentials/users/register"
        uri += "?pin=" + str(pin)
        if name:
            uri += "&name=" + str(name)
        return self.request.get(uri, self.timeout)

    def reset_user(self, pin):
        """
        Register user credentials with PIN number 
        
        :param int pin: PIN number
        
        """
        uri = self.gateway + "/credentials/users/reset"
        uri += "?pin=" + str(pin)

        return self.request.get(uri, self.timeout)

    def get_groups(self, name_filter=None, key=None):
        """
        Get group credentials
        
        :param str filter: group name filter
        :param str key: group key  
        """
        uri = self.gateway + "/credentials/groups"
        if key:
            uri += "?key=" + str(key)
        elif name_filter:
            uri += "?filter=" + str(name_filter)

        return self.request.get(uri, self.timeout)

    def get_group_members(self, name_filter=None, key=None):
        """
        Get group credentials
        
        :param str filter: group name filter
        :param str key: group key  
        """
        uri = self.gateway + "/credentials/groups/members"
        if key:
            uri += "?key=" + str(key)
        elif name_filter:
            uri += "?filter=" + str(name_filter)

        return self.request.get(uri, self.timeout)

    def set_group(self, value):
        """
        Set group credentials 
        
        :param str value: group json data
        """
        uri = self.gateway + "/credentials/groups"

        return self.request.post(uri, json.dumps(value), self.timeout)

    def delete_group(self, name_filter, key=None):
        """
        Delete group credentials
        
        :param str filter: group name filter
        :param str key: group key       
        """
        uri = self.gateway + "/credentials/groups"
        if key:
            uri += "?key=" + str(key)
        elif name_filter:
            uri += "?filter=" + str(name_filter)

        return self.request.delete(uri, self.timeout)


class Device:
    """Represents one Device from a Group

    :param group: an gateway.Group reference or string containing url to a gateway
    :param str name: name of the device returned by the get_device_list(),
                       such as 'device0' root designation or a full name
    :param bool unit: indicates if unit for params is to be fetched.
    :param bool advanced: fetches advanced params of the device if True
    :param bool hidden: fetches hidden params of the device if True
    :param bool development: fetches dev params of the device if True    
    """

    def __init__(self, group = None, name = None, exclude=None, unit=True, accuracy=True, advanced=False, development=False, hidden=False, brand_filter=None):

        if isinstance(group, str):
            group = Group(group)
        if not group:
            group = Group()

        if not name:
            if exclude != None:
                for dev in group.get_device_list(brand_filter):
                    onlist = False
                    for e in exclude:
                        if isinstance(e, Device):
                            if e.name == dev['name'].replace(" ", "_"):
                                onlist = True
                        elif isinstance(e, str):
                            if e.replace(" ", "_") == dev['name'].replace(" ", "_"):
                                onlist = True

                    if not onlist:
                        name = dev['name']
                        break

            else: # take first
                name = group.get_device_list(brand_filter)[0]['name']

        if name == None:
            raise GatewayError('Cannot find an active device.')

        self.group = group
        self.name = name.replace(" ", "_")
        self.unit = unit
        self.accuracy = accuracy
        self.advanced = advanced
        self.hidden = hidden
        self.development = development
        self.active = True
        self.request = group.request
        self.timestamp = 0
        try:            
            status = self.get_device_status()
            if 'name' not in status or status['name'] is None:
                raise ValueError("device " + self.name + " not yet loaded")
            self.name = str(status['name']).replace(" ", "_")
        except Exception as __:
            for i in group.get_device_list(inactive=True):
                if i['name'].replace(" ", "_") == self.name:
                    self.active = False
                    print('Warning: Device is currently inactive but have records.')
                    return
            raise ValueError("device " + self.name + " not available")

    def get_url(self):
        """
        Get device url
        """
        return self.group.gateway_url + "/" + self.name

    def time(self):
        """
        Get last updated time

        :returns: device status in json form
        """
        return self.timestamp

    def get(self, parameter='', after=None, until=None, options=False, attributes=None, args=None):
        """Retrieves one or more parameters of the device.
        If parameter is specified, then only that param info is fetched.
        Otherwise, all the params are fetched.
        A parameter can have a finite range of values, which can be fetched with ``options=True`` flag.

        :param str parameter: name of the parameter
        :param float after: wait until the given UTC time or 'now', and then issue query request
        :param float until: wait for a new sample until given UTC time, and if still not given issue query
        :param bool options: flag to determine if param options are to be fetched
        :param str attributes: fetches attributes of the passed value
        :returns: list of values
        :rtype: list
        :raises: HTTPError, socket.timeout
        """
        uri = self.group.gateway_url + "/"
        uri += self.name + "/"
        uri += parameter.replace('.', '/') + ".json"

        qp = []
        if self.unit:
            qp.append('unit=true')
        if self.accuracy:
            qp.append('accuracy=true')
        if after:
            qp.append('after=' + str(after))
        if until:
            qp.append('until=' + str(until))
        if self.advanced:
            qp.append('advanced=true')
        if self.hidden:
            qp.append('hidden=true')
        if self.development:
            qp.append('development=true')
        if options:
            qp.append('options=true')
        if attributes:
            qp.append('attributes=' + str(attributes))
        if args:
            qp.append('args=true')

        qs = '?' + '&'.join(qp) or ''
        uri += qs
        return self._parse_status(self.request.get(uri, self.group.timeout))

    def get_value(self, parameter, after=None, until=None):
        """Fetches value of the parameter from Device.

        :param str parameter: name of the parameter whose value is to be fetched
        :returns: value of the parameter
        :raises: HTTPError, socket.timeout
        """
        param = parameter.replace('.', '/')
        result = self._parse_time(self.get(param, after, until))[param.split('/')[-1]]
        try:
            return ufloat(float(result['value']), float(result['accuracy']))
        except:
            return result['value']

    def collect(self, parameter, Ts, t_start=None, N=0):
        """Collect N-values of the parameter at given time interval Ts

        :param str parameter: name of the parameter to be fetched
        :param float Ts: time interval between samples in second, or as tupple to define time interval in which sample is expected to arrive
        :param float t_start: start time of first sample to be taken in UTC or None
        :param int N: number of samples to be fetched or 0 for infinite
        :returns: yields (generates) an output to be used with isotel.idm.signal 
        """
        t0 = ts = self.group.time() if t_start == None else t_start
        unit = ''
        try:
            unit = ' [' + self.get(parameter)[parameter]['unit'] + ']'
        except:
            pass
            
        yield (('t [s]', (parameter + unit,)),)
        while True:
            if type(Ts)==tuple:
                value = self.get_value(parameter, after=ts+Ts[0], until=ts+Ts[1])
            else:
                value = self.get_value(parameter, after=ts)
            try:
                valf = value.n
            except:
                valf = float(value)

            if type(Ts)==tuple:
                ts = self.timestamp
                yield ((ts-t0, (valf,)),)
            else:
                yield ((ts-t0, (valf,)),)
                ts += Ts

            if N > 0:
                N -= 1
                if N == 0:
                    break
    
    def __getitem__(self, parameter):
        return self.get_value(parameter)

    def get_records(self, parameter, time_from=None, time_to='last',
                    limit=None, scale=None, average=None, calc=None, show=None):
        """Retrieves values from record storage.

        time_from provides filter value either as a UTC timestamp
        or as 'first' or 'start' strings to indicate that
        retrieval is to be made from first available record.

        time_to provides filter value either as a UTC timestamp
        or one of 'last' or 'end' strings to indicate that
        retrieval is to be made till last n=limit available records.

        Both scale and limit args limit the number of records being returned.
        However, limit returns records each containing ``N`` property.
        scale arg returns records, each containing ``M`` property.
        They are *mutually exclusive*; providing them both raises ``ValueError``
        with appropriate message.
        If none are provided, then gateway restricts the number of records by 100
        with each record containing ``N`` property by default.

        average arg can be used only with scale;
        providing average without scale raises ValueError.
        average specifies the number of records to be averaged on
        to determine a single scale value.
        average usage is hinted by ``SD`` property present in the response.

        :param str parameter: parameter name or stringified number
        :param int or str time_from: time from which records are to be fetched
        :param int or str time_to: time till which records are to be fetched
        :param int limit: max number of records to return with ``N`` property in each sample
        :param int scale: max number of records to return with ``M`` property in each sample
        :param int average: number of records to average on in case of scale

        :returns: list of JSON objects
        :rtype: list
        :raises: HTTPError, socket.timeout, ValueError
        """
        if limit and scale:
            raise ValueError('Provide either limit or scale arg. They are mutually exclusive.')
        if average and not scale:
            raise ValueError('average arg can only be used with scale arg.')

        uri = self.group.gateway_url + "/"
        uri += self.name + "/"
        uri += parameter.replace('.', '/') + ".json"

        qp = []
        if limit:
            qp.append('limit=' + str(limit))
        if scale:
            qp.append('scale=' + str(scale))
        if average:
            qp.append('average=' + str(average))
        if time_from:
            qp.append('from=' + str(time_from))
        if time_to:
            qp.append('to=' + str(time_to))
        if calc:
            qp.append('calc=' + str(calc))
        if show:
            qp.append('show=' + str(show))
        if self.advanced:
            qp.append('advanced=true')
        if self.development:
            qp.append('development=true')
        if self.unit:
            qp.append('unit=true')
        qs = '?' + '&'.join(qp) or ''
        uri += qs
        return self.request.get(uri, self.group.timeout)

    def get_args(self, msg):
        """
        Get arguments for given message
        """
        uri = self.group.gateway_url + "/" + self.name + "/" + str(msg)
        return self._parse_time(self.request.get(uri, self.group.timeout))

    def set_args(self, msg, data):
        """
        Set arguments for given message
        """
        uri = self.group.gateway_url + "/" + self.name + "/" + str(msg)
        return self._parse_time(self.request.post(uri, json.dumps(data), self.group.timeout))

    def set(self, parameter, data, after=None):
        """Set device variables

        :param str parameter: parameter name or stringified number
        :param str data: variable data in json format

        :returns: result of the set values in json format
        :raises: HTTPError, socket.timeout
        """
        uri = self.group.gateway_url + "/"
        uri += self.name + "/"
        uri += parameter.replace('.', '/')

        struct = data.copy()
        if after:
            struct["$after"] = after
        if self.advanced:
            struct['$advanced'] = True
        if self.development:
            struct['$development'] = True

        resp = self.request.post(uri, json.dumps(struct), self.group.timeout)
        return self._parse_status(resp)

    def set_value(self, parameter, data, after=None):
        param = parameter.replace('.', '/')
        result = self.get(param)
        token = param.split('/')[-1]
        result[token]['value'] = data
        result = self._parse_time(self.set(param, result, after))
        try:
            return result[token]['value']
        except:
            return result['status']

    def __setitem__(self, parameter, value):
        return self.set_value(parameter, value)

    def query(self, parameter):
        """
        query given parameter
        :param parameter: 
        :returns: updated latest value of the parameter
        """
        return self.get_value(parameter, after='now')

    def get_device_status(self):
        """
        Get connected device status
       
        :returns: device status in json form
        """
        uri = self.group.gateway_url + "/" + self.name + "/device_status"

        return self.request.get(uri, self.group.timeout)

    def get_protocols(self):
        """
        Get devices protocol tree

        :returns: protocol tree in json form
        """
        port = self.get_device_status()['port']
        return self.group.get_protocols(port)[port]

    def get_stream_hostport(self, streamname):
        """
        Get stream tcp/ip address

        :param streamname: name of the user/custom stream as 'Stream1' or 'Stream2'
        :returns: a tuple (hostname, port)
        """
        return (self.group.get_gatewayname(), int(self.get_protocols()[streamname]['status']['port']))

    def set_device_setting(self, data):
        """
        Set device parameter
                
        :param data: data format {"setting_name":"value"}
        :returns: operation result
        """
        uri = self.group.gateway_url + "/" + self.name + "/device_status"
        return self.request.post(uri, json.dumps(data), self.group.timeout)

    def _parse_status(self, result):

        if isinstance(result, list):
            return result

        status = 'OK'
        try:
            status = result['status']
        except:
            pass

        self._parse_time(result)
        if status.upper() not in ['OK', 'ACK']:
            report = str(status)
            if 'message' in result and result['message'] is not None:
                report = " - ".join([report, str(result['message'])])
            raise IOError(report)
        return result

    def _parse_time(self, result):
        if 'time' in result:
            self.timestamp = float(result['time'])

        return result


class Parameter:
    """Represents a Parameter of a Device

    :param Device device: a device object
    :param str parameter: parameter name or stringified number
    """

    def __init__(self, device, parameter):
        self.device = device
        self.parameter = parameter.replace('.', '/')
        self.timestamp = 0
        self.lunit = None
        self.lvalue = None
        self.struct = self.get()
        param = self.parameter.split('/')[-1]  # Only the last parameter is of interest.

        for v in self.struct.keys():
            if param in v:
                self.key = v

    def time(self):
        return self.timestamp

    def value(self):
        return self.lvalue

    def unit(self):
        return self.lunit

    def get_value(self, after=None):
        value = self._parse_time(self.get(after))
        return value[self.key]['value']

    def get_options(self):
        """Retrieves the list of possible values that the Parameter instance can possess.
        If such a range is defined, then it is returned else None is returned.

        :returns: list of option values else None
        """
        resp = self._parse_time(self.get(options=True))
        return resp[self.key].get('options', None)

    def get(self, after=None, options=False, attributes=None):
        resp = self.device.get(self.parameter, after=after,
                               options=options, attributes=attributes)
        return self._parse_status(resp)

    def get_records(self, time_from=None, time_to='last',
                    limit=None, scale=None, average=None):
        resp = self.device.get_records(self.parameter,
                                       time_from=time_from, time_to=time_to,
                                       limit=limit, scale=scale, average=average)
        return self._parse_status(resp)

    def set_value(self, new_value, after=None):

        self.struct[self.key]['value'] = new_value
        self._parse_time(self.set(self.struct, after))
        return self.value()

    def set(self, data, after=None):
        """Set values of a given parameter(s)

        :param str data: variable data in json format

        :returns: list of values
        :raises: HTTPError, socket.timeout
        """
        resp = self.device.set(self.parameter, data, after)
        return self._parse_status(resp)

    def query(self):
        """
        Send message query for given parameter
        :param data: 
        :return: 
        """
        resp = self.device.query(self.parameter)
        return self._parse_status(resp)

    def _parse_status(self, result):

        if isinstance(result, list):
            return result

        status = 'OK'
        try:
            status = result['status']
        except:
            pass
            # print("Return struct is missing status")
        self._parse_time(result)
        if status.upper() not in ['OK', 'ACK']:
            report = str(status)
            if 'message' in result and result['message'] is not None:
                report = " - ".join([report, str(result['message'])])
            raise IOError(report)

        try:
            self.lvalue = result[self.key]['value']
        except:
            self.lvalue = 'OK'

        try:
            self.lunit = result[self.key]['unit']
        except:
            self.lunit = None

        return result

    def _parse_time(self, result):

        if 'time' in result:
            self.timestamp = float(result['time'])

        return result


class Service:
    """
    An IDM Service instance
    """

    def __init__(self, gateway, service):
        self.gateway = gateway
        self.service = service
        self.service_uri = self.gateway.gateway_url + "/services/" + service
        self.timeout = self.gateway.timeout
        self.request = gateway.request

    def stop(self):
        """
        Send stop signal to service
        """
        data = {"action": "stop"}

        return self.request.post(self.service_uri, json.dumps(data), self.timeout)

    def get_service_data(self, uri):
        """
        Get service data
        
        :param str uri: service uri
        """
        uri = self.service_uri + "/" + uri
        return self.request.get(uri, self.timeout)

    def post_service_data(self, uri, data):
        """
        Post service data
    
        :param str uri: service uri
        :param str data:
        """
        uri = self.service_uri + "/" + uri
        return self.request.post(uri, json.dumps(data), self.timeout)

    def get_service_property_list(self):
        """Retrieves a list of properties related to the service.
        :returns: list of property names
        :rtype: list
        """

        uri = self.service_uri + "/properties" + '.json'
        return self.request.get(uri, self.timeout)

    def get_service_property(self, prop, params=None):
        """Fetches value of *prop* property.
        The object returned has a definite structure:
        ``{ prop: {data: <value>}, time: <tstmp> }``

        :param str identifier: property group identifier
        :param str prop: name of the property
        :returns: value of the property
        :rtype: dict
        """
        prop = prop.replace(".", "/")

        uri = self.service_uri + "/properties/" + prop + '.json'

        uri = self.gateway.add_params_to_uri(uri, params=params)
        return self.request.get(uri, self.timeout)

    def set_service_property(self, prop, value, append=False):
        """Sets or overwrites the value of the *prop* property with *val* value.
        The *val* passed is stored and enclosed in:
        ``{prop: {data: <val>}}``

        :param str identifier: property group identifier
        :param str prop: name
        :param str|dict|number value: value
        :returns: result with structure ``{result: Error|OK}``
        :rtype: dict
        """
        prop = prop.replace(".", "/")
        uri = self.service_uri + "/properties/" + prop + '.json'
        if append:
            uri += '?append=true'

        return self.request.post(uri, json.dumps(value), self.timeout)

    def delete_service_property(self, prop, params=None):
        """Deletes the `prop` property.
        The entire property, not only its value, is removed.
        
        :param str identifier: property group identifier
        :param str prop: name
        :return: response with structure: ``{result: Error|OK}``
        :rtype: dict
        """
        prop = prop.replace(".", "/")

        uri = self.service_uri + "/properties/" + prop + '.json'

        uri = self.gateway.add_params_to_uri(uri, params=params)
        return self.request.delete(uri, self.timeout)



# Major number means incompatible API
# Minor number means new add-ons but backward compatibility
# Patch number means bugfixes or other improvements that do not affect API
#   with development stage either alpha: a1,a2,a3,.. or beta: b1,b2,b3
__version__ = '1.2'



import urllib.request
import json
import gzip
import ssl
import http.client
import base64
import string
import datetime
import time

from datetime import timezone

from urllib.parse import urlparse
from urllib.parse import quote
from http.client import HTTPSConnection
from http.client import HTTPConnection
from http.client import HTTPException
from http.client import BadStatusLine


def get_context(untrusted=True, cert_file=None, cert_pass=None):
    """Decodes resp. If the decoded resp is JSON, then a JSON object is returned.

    :param bytes resp: Encoded response returned by request.
    :returns: decoded response or parse JSON object if response is JSON.
    """
    ctx = ssl.create_default_context()

    if untrusted:
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

    if cert_file is not None:
        ctx.load_cert_chain(cert_file, password=cert_pass)

    return ctx


def encode_auth_header(username, password):
    """
    Encodes username and password for HTTP asic authentication
    """
    encoded = base64.b64encode(bytes(username, 'UTF-8') + b':' + bytes(password, 'UTF-8'))
    header = {'Authorization': b'Basic ' + encoded}
    return header


def _decode_resp(response):
    """Unpacks response

    :param bytes resp: Encoded response returned by request.
    :returns: decoded response or parse JSON object if response is JSON.
    """
    response_headers = response.info()
    encoding = response_headers['Content-Encoding']
    r = response.read()
    if encoding is not None and str(encoding) == "gzip":
        try:
            r = gzip.decompress(r)
        except OSError:
            pass
    return _parse_resp(r)


def _parse_resp(resp):
    """Decodes resp. If the decoded resp is JSON, then a JSON object is returned.

    :param bytes resp: Encoded response returned by request.
    :returns: decoded response or parse JSON object if response is JSON.
    """
    parsed = resp.decode('utf-8')
    try:
        return json.loads(parsed)
    except:
        return parsed


class RequestHandler:

    def __init__(self, ssl_context, header, keep_alive=True, keep_alive_timeout=10):
        """
        Initialize request handler
        """
        self.ssl_context = ssl_context
        self.header = header
        self.keep_alive = keep_alive
        self.keep_alive_timeout = keep_alive_timeout
        self.connection = None
        self.conn_expired = datetime.datetime.now(timezone.utc)
        self.finished_req = True

    def init_connection(self, uri_netloc, timeout):
        """
        Initialize connection if necessary
        :param uri_netloc str: net location
        :returns: True if connection is being reused, False otherwise
        """
        now = datetime.datetime.now(timezone.utc)

        if not self.connection or now > self.conn_expired or not self.keep_alive or not self.finished_req:
            self.close_connection()
            if self.ssl_context:
                self.connection = HTTPSConnection(uri_netloc, timeout=timeout, context=self.ssl_context)
            else:
                self.connection = HTTPConnection(uri_netloc, timeout=timeout)
            return False
        else:
            return True

    def close_connection(self):

        if self.connection:
            self.connection.close()
            self.connection = None

    def connection_needs_closing(self, response):
        """
        Checks header to see if connection needs to be closed
        :param bytes resp: Encoded response returned by request.

        """
        response_headers = response.info()
        con = response_headers['Connection']

        if not con:
            con = response_headers['connection']

        if con and str(con).startswith("close"):
            return True

        return False

    def get(self, url, timeout):
        """Makes a GET request on url.
        The function waits for response till timeout is reached.

        :param str url: Absolute URL to executed the GET request
        :param int timeout: Time till which to wait for response
        :param bool untrusted: request lets true untrusted ssl certificate
        :returns: decoded response. If response is JSON, then JSON object is returned.

        """
        return self._handle_request("GET", url, timeout=timeout)

    def post(self, url, data, timeout):
        """Makes a POST request on url.
        data is sent with the request. data is expected to be a string of JSON object.
        The function waits for response till timeout is reached.

        :param str url: Absolute URL to executed the GET request
        :param str data: stringified JSON data
        :param int timeout: Time till which to wait for response
        :param bool untrusted: request lets true untrusted ssl certificate
        :returns: decoded response. If response is JSON, then JSON object is returned.
        """
        return self._handle_request("POST", url, data=data, timeout=timeout)

    def delete(self, url, timeout):
        """Makes a DELETE request on url.
        The function waits for response till timeout is reached.

        :param str url: Absolute URL to executed the GET request
        :param int timeout: Time till which to wait for response
        :param bool untrusted: request lets true untrusted ssl certificate
        :returns: decoded response. If response is JSON, then JSON object is returned.
        """
        return self._handle_request("DELETE", url, timeout=timeout)

    def _send_request(self, conn, method, path, headers, data=None):

        if method == "POST":
            conn.request(method, path, data, headers=headers)
        else:
            conn.request(method, path, headers=headers)

        return conn.getresponse()

    def _handle_request(self, method, url, data=None, timeout=10):
        """
        Handle request with low level library to retain version compatibility
        """
        uri_p = urlparse(url)
        path = uri_p.path
        path = quote(path)

        reuse = self.init_connection(uri_p.netloc, timeout)
        self.finished_req = False
        headers = {'Accept-Encoding': 'gzip'}

        if method == "POST":
            headers['content-type'] = 'application/json'

        if self.header:
            for key, value in self.header.items():
                headers[key] = value

        if len(uri_p.query) > 0:
            path += '?' + uri_p.query

        try:
            res = self._send_request(self.connection, method, path, headers, data=data)
        except Exception as e:
            if reuse:
                self.close_connection()
                self.init_connection(uri_p.netloc, timeout)
            time.sleep(5)
            res = self._send_request(self.connection, method, path, headers, data=data)

        self.conn_expired = datetime.datetime.now(timezone.utc) + datetime.timedelta(seconds=self.keep_alive_timeout)

        if self.connection_needs_closing(res):
            self.close_connection()

        if res.status >= 400:
            msg = _decode_resp(res)
            try:
                msg = str(msg['message'])
            except:
                msg = str(msg)
            raise HTTPException(str(res.status) + " " + str(res.reason) + ": " + msg)

        fin = _decode_resp(res)
        self.finished_req = True
        return fin



"""ISOTEL IDM Base Representation of Signals and their Generators

Structure:
 - values is an array of measurement values
 - signal is a dict sorted by channel each having a pair of lists: y and x values
 - stream is a tuple of individual measurements from multiple multi-channel signals
 - channel refers to a measurement input by name

"""
from math import *
import re
import time

try:
    if get_ipython:
        from ipywidgets import interact, IntSlider, ToggleButton
        from bokeh.io import push_notebook, show, output_notebook
        from bokeh.plotting import figure
        from bokeh.models import Range1d
except:
    pass

try:
    # Needed by signal2spectrum
    from scipy.fftpack import fft
except:
    pass

try:
    # Needed by SNR
    import scipy.optimize
    import numpy as np
except:
    pass

try:
    import pyqtgraph as pg
except:
    pass

def search_edge(values, threshold, hysteresis=0, edge='rising', start_idx=0):
    state = None
    oldstate = None
    thr_low = threshold - hysteresis
    thr_high= threshold + hysteresis

    for i,v in enumerate(values):
        if v < thr_low:
            state = 0
        elif v > thr_high:
            state = 1

        if oldstate != None and state != None and i >= start_idx:
            if edge == 'rising' or edge == 'any':
                if oldstate < state:
                    return i
            if edge == 'falling' or edge == 'any':
                if oldstate > state:
                    return i

        oldstate = state

    return None

def tofloat(values, index):
    if index:
        return float(values[index])
    return float('nan')

def get_stream_bychannel(header, name):
    for i,s in enumerate(header):
        for j,ch in enumerate(s[1]):
            if ch.startswith(name):
                return lambda s: s[i][1][j]

def replace_channels2streams(header, expr):
    expr_out = expr
    for token in re.findall(r"[A-Z]+\d*", expr): # Assumes channel names to be uppercase followed by optional numbers
        for i,s in enumerate(header):
            for j,ch in enumerate(s[1]):
                if ch.startswith(token):
                    expr_out = expr_out.replace(token, 's['+str(i)+"][1]["+str(j)+"]")
    return expr_out

def trigger(stream, cond, precond=None, P=50, N=50, single_shot=False, post_delay=0):
    try:
        hdr = next(stream)
        yield hdr

        precond_triggered = False
        triggered = False
        prebuf = [[]]*P
        prebuf_wi = 0
        prebuf_size = 0
        N_preset = N
        trigger_ts = 0

        exec('global trigger_cond; trigger_cond = lambda s:' + replace_channels2streams(hdr, cond))
        if precond:
            exec('global pretrigger_cond; pretrigger_cond = lambda s:' + replace_channels2streams(hdr, precond))
        else:
            precond_triggered = True

        for s in stream:
            if not triggered and prebuf_size >= P:
                ts = time.time()
                if ts > trigger_ts:
                    if precond and not precond_triggered:
                        if pretrigger_cond(s) > 0:
                            precond_triggered = True

                    if precond_triggered > 0 and trigger_cond(s) > 0:
                        triggered = True
                        trigger_ts = ts + post_delay
                        if P >= 0:
                            for __ in range(P):
                                if prebuf[prebuf_wi] != []:
                                    yield prebuf[prebuf_wi]
                                prebuf_wi = (prebuf_wi+1) % P

            if triggered:
                N -= 1
                if N >= 0:
                    yield s
                elif not single_shot:
                    precond_triggered = False
                    triggered = False
                    N = N_preset
                else:
                    return

            # Prebuffer keeps filling for subsequent shots
            if P > 0:
                prebuf[prebuf_wi] = s
                prebuf_wi = (prebuf_wi+1) % P
                prebuf_size += 1
    finally:
        stream.close()

def addmath(stream, name, expr, remove_NaN=True):
    try:
        hdr = next(stream)
        yield hdr + ( ('t [s]', (name,)), )

        exec('global math_expr; math_expr = lambda s:' + replace_channels2streams(hdr, expr))
        for s in stream:
            t = None
            for st in s:
                if st[0] != None:
                    t = st[0]
                    break
            if t:
                try:
                    result = math_expr(s)
                    if (remove_NaN == True and result == result) or remove_NaN == False:
                        yield s + ( (t, (result,)), )
                except:
                    pass
    finally:
        stream.close()

def merge(*streams):
    try:
        for s in zip(*streams):
            merged = ()
            for t in s:
                for c in t:
                    merged += (c,)
            yield merged
    finally:
        for s in streams:
            s.close()

def mergeNsync(*streams):
    try:
        ss = [s for s in streams]

        # Merge Header
        v = [next(s) for s in ss]
        merged = ()
        for t in v:
            for c in t:
                merged += (c,)
        yield merged

        # Sync and Merge Readings
        v = [next(s) for s in ss]
        loop = True
        while loop:
            t_min = v[0][0][0]
            for t in v:
                t_min = t[0][0] if t[0][0] < t_min else t_min

            merged = ()
            for i,t in enumerate(v):
                used = False
                for c in t:
                    if c[0] != None and c[0] <= t_min:
                        merged += (c,)
                        used = True
                    else:
                        merged += ((None, (None,)),)
                if used:
                    try:
                        v[i] = next(ss[i])
                    except:
                        loop = False
            yield merged
    finally:
        for s in streams:
            s.close()

def replay(stream):
    for s in stream:
        yield s

def to_tsv(stream):
    def convert(num):
        if num != None and num == num:
            return str(num)
        return ''

    def make_tsv_line(tup):
        line = ''
        for s in tup:
            line += convert(s[0]) + '\t'
            for y in s[1]:
                line += convert(y) + '\t'
        return line

    try:
        for s in stream:
            yield make_tsv_line(s)
    finally:
        stream.close()

def stream2signal(stream, split=None, relative_time=False, incremental=False):
    try:
        header = next(stream)
        i = 0
        out = None
        for src in stream:
            i += 1
            if i == 1:
                out = {}
                x_ref = 0
                for s in header:
                    for a in s[1]:
                        if a != None:
                            out[str(a)] = dict({'y': [], 'x': []})

            # Collect values, for each stream in src
            for s,h in zip(src, header):
                for a,v in zip(h[1],s[1]):
                    if s[0] != None:
                        out[str(a)]['y'] += [ v ]
                        if relative_time and i == 1:
                            x_ref = s[0]

                        out[str(a)]['x'] += [ s[0] - x_ref ]

            if i == split:
                yield out
                out = None
                i = 0
            elif incremental:
                yield out

        # Flush remaining
        if out != None:
            yield out
    finally:
        stream.close()

def rms(y, ac=False):
    if ac:
        return rms(np.array(y) - np.mean(y))
    return np.sqrt( np.mean(np.array(y)**2) )

def snr_dB(signal, noise):
    return 20*log10(rms(signal) / rms(noise))

def sinad_dB(signal):
    """
    Calculates SINAD by fitting a single ideal tone sine to the signal
    to be subtracted from the actual signal.

    :param signal: a signal dict providing x and y arrays
    :returns: SNR in dB and frequency
    """
    def sine_fitter(x, y):
        def sinfun(t, A, w, p, c):
            return A * np.sin(w*t + p) + c

        ff = np.fft.fftfreq(len(x), (x[1]-x[0]))
        Fyy = abs(np.fft.fft(y))
        guess_freq = abs(ff[np.argmax(Fyy[1:])+1])
        guess_amp = np.std(y) * 2.**0.5
        guess_offset = np.mean(y)
        guess = np.array([guess_amp, 2.*np.pi*guess_freq, 0., guess_offset])
        popt, pcov = scipy.optimize.curve_fit(sinfun, x, y, p0=guess)
        A, w, p, c = popt
        return sinfun(x, A, w, p, c), w/(2.*np.pi), abs(A)

    x = np.array(signal['x'])
    y = np.array(signal['y'])
    y_ideal, f, A = sine_fitter(x, y)
    return snr_dB(y, y - y_ideal), f, A

def enob(sinad):
    return (sinad - 1.76) / 6.02

def blackmanharris(N, sym, a):
    k = np.linspace(0,len(a)-1,len(a))
    w = np.zeros(N)
    M = N-1 if sym else N
    for i in range(N):
        w[i] = np.sum( a * np.cos(2*np.pi*k*i/M) )
    return w

def blackmanharris3(N, sym=True):
    BLACKMANHARRIS_3TH = [0.42,  -0.5, 0.08]
    return blackmanharris(N, sym, BLACKMANHARRIS_3TH)

def blackmanharris4(N, sym=True):
    # https://liquidsdr.org/doc/windowing/
    BLACKMANHARRIS_4TH = [0.35875, -0.48829, 0.14128, -0.01168]
    return blackmanharris(N, sym, BLACKMANHARRIS_4TH)

def blackmanharris7(N, sym=True):
    # https://dsp.stackexchange.com/questions/51095/seven-term-blackman-harris-window
    BLACKMANHARRIS_7TH = [0.27105140069342, -0.43329793923448, 0.21812299954311,
        -0.06592544638803, 0.01081174209837, -0.00077658482522, 0.00001388721735]
    return blackmanharris(N, sym, BLACKMANHARRIS_7TH)

def tospectrum(signal, log_ref=1, log_unit='dB', start_fi=0, window=blackmanharris3):
    try:
        Nw = None
        for shot in signal:
            fft_shot = {}
            for s in shot:
                N = len(shot[s]['y'])
                if Nw != N:
                    w = window(N, False) if window else 1
                    w = w/rms(w)
                    Nw = N
                Y = [20*log10(2*abs(i+1e-99)/log_ref/N) for i in fft(w*shot[s]['y'])]
                dT = (shot[s]['x'][-1] - shot[s]['x'][0]) / (N-1)
                X = [i / ((N-2)*dT) for i in range(N//2)]
                fft_shot[s.replace('[', '['+log_unit)] = { 'y': Y[start_fi:N//2], 'x' : X[start_fi:] }

            yield fft_shot

    except KeyboardInterrupt:
        pass
    finally:
        signal.close()

    return fft_shot


DISPLAY_MODE_AVG = 'avg'
DISPLAY_MODE_PEAK = 'peak'
SCOPE_CH_COLORS = ['darkorange', 'darkmagenta', 'darkcyan', 'darkred', 'darkgreen', 'red', 'blue', 'brown']

def scope(signal, width=700, height=400, title="Unnamed",
          colors = SCOPE_CH_COLORS, y_range = None,
          x_label = 't [s]', skip=0, post_func=None, display_mode=None, avg_k=0.5):
    
    assert(avg_k<=1)
    try:
        get_ipython
        output_notebook()
    except:
        print('Runs inside jupyter with bokeh plots only')
        return

    def plot_lines(plot, shot):
        lines  = []
        for i, l in enumerate(shot):
            lines += [ plot.line(x = shot[l]['x'], y = shot[l]['y'], color = colors[i], legend = l) if colors[i] else None ]

        return lines

    def plot_setup():
        plot = figure(title=title, x_axis_label=x_label,
                #y_axis_label=   Common unit on Y axis can cause many issues, it's been added to the legent
                plot_width=width, plot_height=height, background_fill_color='white', border_fill_color='white',
                tools="pan,crosshair,box_zoom,reset,save,wheel_zoom,hover")

        plot.xgrid.grid_line_color = 'lightgray'
        plot.ygrid.grid_line_color = 'lightgray'
        if y_range:
            plot.y_range = Range1d(y_range[0], y_range[1])

        return plot

    plot = plot_setup()
    shot = last_valid_shot = None
    plot_handle = None

    try:
        i = 0
        valid = True
        for shot in signal:
            if plot_handle == None:
                lines = plot_lines(plot, shot)
                last_valid_shot = shot
                plot_handle = show(plot, notebook_handle=True)
            else:
                i += 1
                if i > skip:
                    for i, l in enumerate(shot):
                        if lines[i]:
                            # Only update screen with the same number of samples
                            if len(lines[i].data_source.data['y']) == len(shot[l]['y']):
                                valid = True
                                if display_mode == 'avg':
                                    avg = [a*(1-avg_k)+b*avg_k for a,b in zip(lines[i].data_source.data['y'],shot[l]['y'])]
                                    lines[i].data_source.data['y'] = avg
                                    shot[l]['y'] = avg
                                elif display_mode == 'peak':
                                    peak = [max(a,b) for a,b in zip(lines[i].data_source.data['y'],shot[l]['y'])]
                                    lines[i].data_source.data['y'] = peak
                                    shot[l]['y'] = peak
                                else:
                                    lines[i].data_source.data = shot[l]
                            else:
                                valid = False

                    push_notebook(handle=plot_handle)
                    i = 0

            if valid:
                last_valid_shot = shot
                if post_func:
                    post_func(shot)

    except KeyboardInterrupt:
        pass
    finally:
        signal.close()

    return last_valid_shot


def qscope(signal, width=700, height=400, title="Unnamed",
          colors = SCOPE_CH_COLORS,
          y_range = None, x_label = 't [s]', skip=0, post_func=None, display_mode=None, avg_k=0.5, chain=False):

    assert(avg_k<=1)

    def plot_lines(plot, shot):
        lines  = []
        for i, l in enumerate(shot):
            lines += [ plot.line(x = shot[l]['x'], y = shot[l]['y'], color = colors[i], legend = l) if colors[i] else None ]

        return lines

    def plot_setup():
        qp = pg.plot(title=title)
        qp.setLabel('bottom', text = x_label.split('[')[0], units = x_label.split('[')[1].split(']')[0])
        qp.showGrid(True,True,0.7)
        qp.showButtons()
        qp.addLegend(offset=(-5,5))
        return qp
    
    qp = plot_setup()
    shot = last_shot = None
    pen = [pg.mkPen(pg.QtGui.QColor(color)) for color in colors]

    try:
        i = 0
        valid = True
        legend = True
        for shot in signal:
            i += 1
            if i > skip:
                for i, l in enumerate(shot):
                    if last_shot:
                        # Only update screen with the same number of samples
                        if len(last_shot[l]['y']) == len(shot[l]['y']):
                            valid = True
                            if display_mode == 'avg':
                                avg = [a*(1-avg_k)+b*avg_k for a,b in zip(last_shot[l]['y'],shot[l]['y'])]
                                shot[l]['y'] = avg
                            elif display_mode == 'peak':
                                peak = [max(a,b) for a,b in zip(last_shot[l]['y'],shot[l]['y'])]
                                shot[l]['y'] = peak
                        else:
                            valid = False

                    if valid:
                        qp.plot(shot[l]['x'], shot[l]['y'], pen = pen[i], name = l if legend else None, clear=i==0 )

                i = 0
                legend = False

            if valid:
                last_shot = shot
                if post_func:
                    post_func(shot)

            pg.QtGui.QApplication.processEvents()

            if chain:
                yield shot

    except KeyboardInterrupt:
        pass
    finally:
        signal.close()

    if not chain:
        yield last_shot



import os
import re
from setuptools import setup, find_packages

def extract_version(verfile):
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"

    with open(verfile, "rt") as f:
        data = f.read()
        mo = re.search(VSRE, data, re.M)

    if mo:
        return mo.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s." % (verfile,))

short_desc = ('A python framework for the ISOTEL Data-acquisition & Metering '
              'Smart Gateway for easy integration with Internet of Things.')

long_desc = ('ISOTEL Data-acquisition & Metering Project offers a complete solution '
             'for wired and wireless communication with sensor & control devices. '
             'IDM Project strives for correct presentation of sensor measurements, '
             'accuracy, precision and easy description of their complex transfer '
             'functions to fulfil requirements of accurate metering and precise '
             'control in control systems.\n'
             'Package provides API to IDM Smart Gateway and  in addition hardware '
             'specific support for: Monodaq-U-X, and Energy Control Modules.\n')

setup(
    name='isotel-idm',
    version=extract_version('isotel/idm/_version.py'),
    url='http://isotel.eu/idm',
    author='Isotel Team',
    author_email='pypi@isotel.eu',
    license='LGPL3.0',
    description=short_desc,
    long_description=long_desc,

    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Telecommunications Industry',
        'Intended Audience :: Manufacturing',

        'Topic :: Communications',
        'Topic :: Home Automation',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
        'Topic :: Terminals',
        'Topic :: Terminals :: Serial',

        'Framework :: Flask',
        'Framework :: IPython',
        'Framework :: Jupyter',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        #'Programming Language :: Python :: 2',
        #'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],
    keywords='idm gateway data acquisition metering server scalable sensor network protocols heterogeneous hardware devices DIY',
    packages=find_packages(exclude=['IoT', 'contrib', 'docs', 'tests*', 'notebook-samples', 'examples*', 'isotel/idm/devel']),
    install_requires=[
        'requests>=2.19.1',
        'future>=0.16.0',
        'uncertainties>=3.0.2'
    ],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "httpretty>=0.9.5"],
)



"""ISOTEL IDM support to decode signals from user defined custom streams into IDM Streams
"""
import socket

class CustomStream():
    """Open and Decode Stream from Custom/User Protocol of either endianness and sizes.

    :param stream_hostport: a tuple returned by gateway.Device.get_stream_hostport()
    :param packet_format: specifies one or more variables after the first packet time specs in the format as
           [('t [s]', packet_size, packet_interval), ('val [V]', start_byte, end_byte, scale, post_offset), ..],
           where if start_byte < end_byte denotes little endian, 
           if either start or end byte is negative value determines signed othewise unsigned
    :param start_func: is lambda function executed at start
    :param stop_func: is lambda function executed before starting the transfer and at the end
    :param packet_timeout: denotes tcp/ip socket timeout to retrieve the data
    """
    def __init__(self, stream_hostport, packet_format, start_func, stop_func, packet_timeout=0.1):
        self.packet_format = packet_format
        self.start_func = start_func
        self.stop_func = stop_func

        self.stream_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.stream_socket.connect(stream_hostport)
        self.stream_socket.settimeout(packet_timeout)

    """
    Is a generator yielding the IDM signal.stream compatible output to be used with scope.

    :param N: number of samples (packets) to fetch
    """
    def fetch(self, N=10):
        self.stop_func()
        while True:
            try:
                self.stream_socket.recv(1024)
            except socket.timeout:
                break
        try:
            yield ((self.packet_format[0][0], tuple([i[0] for i in self.packet_format[1:]])),)
            self.start_func()
            for t in range(N):
                pckt = self.stream_socket.recv( self.packet_format[0][1] )
                vals = []
                for i in range(1,len(self.packet_format)):
                    st  = self.packet_format[i][1]
                    end = self.packet_format[i][2]
                    signed = True if st < 0 or end < 0 else False
                    st  = abs(st)
                    end = abs(end)
                    if end > st:
                        vals += [self.packet_format[i][3] * int.from_bytes(pckt[st:end], byteorder='little', signed=signed) - self.packet_format[i][4]]
                    else:
                        vals += [self.packet_format[i][3] * int.from_bytes(pckt[end:st], byteorder='big', signed=signed) - self.packet_format[i][4]]

                yield ((t*self.packet_format[0][2], tuple(vals)),)

            self.stop_func()

        except KeyboardInterrupt:
            self.stop_func()

        except (GeneratorExit, socket.timeout):
            self.stop_func()
            raise
